# Kragen Sitaker, 2001-12-05

# Produce HTML renderings of Python objects.
# For stylesheets: the following classes are predefined:
# py_num --- numbers
# py_str --- strings
# py_left --- left delimiters for lists, tuples, etc.
# py_right --- right delimiters, same
# py_list --- a list
# py_tuple --- a tuple
# py_dict --- a dictionary
# py_dictkey --- a dictionary key
# py_dictvalue --- a dictionary value

# In Python, there are functions 'repr' and 'str' for constructing
# Python expressions and human-readable descriptions of objects,
# respectively.

# 'repr' in particular is used to good effect by the built-in
# read-eval-print loop in Python.  Whatever object was returned by the
# evaluation gets printed out.

# Object instances may have attributes '__repr__' and '__str__' that
# tell repr() and str() how to represent them, and these are often
# recursive, calling repr() or str() on other objects.  The repr() and
# str() of Python's built-in containers includes the repr() and str()s
# of their contents.

# Still, both 'repr' and 'str' are constrained to produce readable
# ASCII text representations of objects.  I propose a third function,
# htmlstr(), which generates a readable HTML representation of an
# object suitable for debugging output.  In environments where using
# htmlstr() is feasible, it can have several advantages over repr()
# and str():
# - far richer formatting is available
# - it doesn't have to be escaped to display as HTML
# - some objects, such as URL objects, may display with links to help
#   elucidate their content
# - other objects may want to display with links to help elucidate their
#   semantics
# - some objects may even want to display as images (sometimes doing this
#   may require some side effect, such as making an image available on an
#   HTTP server somewhere)

# For backward-compatibility, htmlstr() follows the following rules to compute
# the HTML representation of the object:
# - if an object has an htmlstr() method, it will call that method and
#   return its result
# - otherwise, if the object is a class instance, htmlstr() will look for
#   an htmlstr() function registered for that class or any of its base classes,
#   recursively, and run that function on it
# - otherwise, htmlstr() will look for an htmlstr() function registered for
#   that object's type
# - otherwise, htmlstr() will HTML-escape repr(the object).

# You can register htmlstr() functions for types and classes by calling
# register_htmlstr(type_or_class, function).

# Generally, you should import this module after any other modules you
# import.  It checks to see whether a bunch of modules are loaded, and
# if so, it registers htmlstr() functions for data types in them.  (I
# thought it would be inconsiderate to load most of the standard
# library just in case you were using it, and there doesn't seem to be
# a way of being notified when you import modules later.)

import string, types
from types import *

_registered_htmlstr_functions = {}

def _html_escape(str):
    return string.replace(
        string.replace(
        string.replace(
        string.replace(str,
                       '&', '&amp;'),
        '<', '&lt;'),
        '>', '&gt;'),
        '"', '&quot;')

def htmlstr(object):
    if hasattr(object, 'htmlstr') and callable(object.htmlstr):
        return object.htmlstr()
    if hasattr(object, '__class__'):
        classes = [object.__class__]
        while classes:
            if _registered_htmlstr_functions.has_key(classes[0]):
                return _registered_htmlstr_functions[classes[0]](object)
            else:
                classes[0:1] = classes[0].__bases__
    if _registered_htmlstr_functions.has_key(type(object)):
        return _registered_htmlstr_functions[type(object)](object)
    return '<span class="py_fallback">%s</span>' % _html_escape(repr(object))

def register_htmlstr(type_or_class, function):
    _registered_htmlstr_functions[type_or_class] = function

# for builtin types:
# I don't think we can get much better for numeric types than repr(),
# but we can tag them for styling:

def _span(htmlclass, string):
    return '<span class="py_%s">%s</span>' % (htmlclass, string)

def _number_htmlstr(num):
    return _span('num', _html_escape(repr(num)))
for thistype in IntType, LongType, FloatType, ComplexType:
    register_htmlstr(thistype, _number_htmlstr)

# likewise strings:

def _str_htmlstr(string):
    return _span('str', _html_escape(repr(string)))
register_htmlstr(StringType, _str_htmlstr)
if hasattr(types, 'UnicodeType'):
    exec "register_htmlstr(UnicodeType, _str_htmlstr)"

# tuples, lists, and dicts can benefit, though, I think.

def _seqtable(htmlclass, lparen, contents, rparen):
    # XXX is div really kosher anywhere a table is?
    return '''<div class="py_%s"><table cellspacing="0" cellpadding="0"><tr
    ><td valign="top" class="py_left">%s</td><td>%s</td
    ><td valign="bottom" class="py_right">%s</td
    ></tr></table></div>''' % (htmlclass, lparen, contents, rparen)

register_htmlstr(TupleType, lambda tuple:
                 _seqtable('tuple', '(',
                           string.join(map(htmlstr, tuple), ',\n'), ')'))
register_htmlstr(ListType, lambda list:
                 _seqtable('list', '[',
                           string.join(map(htmlstr, list), ',\n'), ']'))
def _dict_htmlstr(dict):
    # this is probably not the easiest or most straightforward way
    # to put commas on all but the last, but it's the first way I thought of
    items = dict.items()
    for ii in range(len(items)-1):
        items[ii] = (htmlstr(items[ii][0]), htmlstr(items[ii][1]) + ",")
    if items:
        items[-1] = map(htmlstr, items[-1])
    # the key must be valigned at the top in case the value is more than
    # one line long (and is more lines than the key), and the value must
    # be valigned at the bottom in case the key is more than one line long
    # (and is more lines than the value).  This still doesn't work when
    # they're both more than one line long.  I don't know how to solve that.
    # <nobreak>?  Ew.  (Strangely enough, at least in Mozilla 0.9.5, the
    # line wrapping doesn't actually seem to break things.)
    contents = [
        '''<tr><td class="py_dictkey" valign="top">%s:</td
        ><td class="py_dictvalue" valign="bottom">%s</td></tr>''' % tuple(item)
        for item in items]
    return _seqtable('dict', '{',
                     '<table>%s</table>' % string.join(contents, ""),
                     '}')
register_htmlstr(DictType, _dict_htmlstr)

# and that pretty much covers the builtin types we can do anything
# sensible with; it might be reasonable to want to cover NoneType,
# because it's so common; XRangeType and SliceType might benefit from
# tagging their begins and ends, and maybe FileType could benefit from
# a hyperlink to the file.  Maybe built-in functions we know of could
# have links to their documentation, with helpful link titles... and
# how about tracebacks?

# But what of types from the library?  Things from weakref, re, array,
# Numpy, mutex, threading, socket, mmap, various dbm libs (anydbm,
# dumbdbm, dbhash, bsddb, dbm, gdbm), zlib, dl, the various network
# libraries (starting with urllib), asyncore, COM, PIL, parse trees?
